#!/usr/bin/env python3

import torch
from torch import Tensor
import torch.onnx

import torch.nn.functional as F

import numpy as np

from pathlib import Path
import sys

import cv2

from torchvision import transforms

from .visualizer import imshow

rtstereo_dir = sys.argv[1]
rtstereo_model = sys.argv[2]

sys.path.insert(0, rtstereo_dir)
from models import RTStereoNet

def b2mb(x): return (x/2**20)

class StereodemoPerformanceMonitor:
    def __init__(self, name, load_model, do_inference, is_gpu: bool):
        self.load_model = load_model
        self.do_inference = do_inference
        self.is_gpu = is_gpu
        self.name = name

    def run (self):
        import time
        model = self.load_model ()
        timings = []
        for i in range (0, 5):
            tstart = time.time ()
            self.do_inference (model)
            tend = time.time ()
            dt = tend - tstart
            timings.append (dt)
            print (f'{dt=}')
        print (f'{self.name}: timings {timings}')
        
        if self.is_gpu:
            import gc
            peak_memory_inference_mb = []
            for i in range (0, 5):
                gc.collect ()
                torch.cuda.empty_cache()
                torch.cuda.reset_max_memory_allocated() # reset the peak gauge to zero
                model = self.load_model ()
                peak_after_load = torch.cuda.max_memory_allocated()
                self.do_inference (model)
                peak_after_inference = torch.cuda.max_memory_allocated()
                print (f'{peak_after_load=}', peak_after_load)
                print (f'{peak_after_inference=}', peak_after_inference)
                peak_memory_inference_mb.append (b2mb(peak_after_inference))
            print (f'{self.name}: peak memory (MB) {peak_memory_inference_mb}')

def save_torchscript(net, output_file, device):
    scripted_module = torch.jit.script(net)
    # net = net.to(device)
    # sample_input = (torch.zeros(1,3,256,256).to(device), torch.zeros(1,3,256,256).to(device))
    # scripted_module = torch.jit.trace(net, sample_input)
    torch.jit.save(scripted_module, output_file)
    return scripted_module

def save_onnx(net, output_file):
    torch.onnx.export(net,                   # model being run
                  sample_input,              # model input (or a tuple for multiple inputs)
                  output_file,               # where to save the model (can be a file or file-like object)
                  export_params=True,        # store the trained parameter weights inside the model file
                  opset_version=11,          # the ONNX version to export the model to
                  do_constant_folding=True,  # whether to execute constant folding for optimization
                  input_names = ['left', 'right'],   # the model's input names
                  output_names = ['disparity'], # the model's output names
                  dynamic_axes={'left' : {0 : 'batch_size', 2 : 'width', 3 : 'height' },  # variable length axes, except channels
                                'right' : {0 : 'batch_size', 2 : 'width', 3 : 'height' },
                                'output' : {0 : 'batch_size', 2 : 'width', 3 : 'height'}})

def show_color_disparity (name: str, disparity_map: np.ndarray):
    min_disp = 0
    max_disp = 64
    norm_disparity_map = 255*((disparity_map-min_disp) / (max_disp-min_disp))
    disparity_color = cv2.applyColorMap(cv2.convertScaleAbs(norm_disparity_map, 1), cv2.COLORMAP_MAGMA)
    imshow (name, disparity_color)

def export_models ():
    checkpoint_file = rtstereo_model
    net = RTStereoNet(maxdisp=192, device='cpu')
    checkpoint = torch.load(checkpoint_file)
    net.load_state_dict(checkpoint['state_dict'])
    net.eval()

    # Hacky way to check the original model and make sure the export
    # is not screwing up the results.
    if False:
        # left = cv2.imread("datasets/drivingstereo/2018-07-11-14-48-52/2018-07-11-14-48-52_2018-07-11-14-50-10-570_left.jpg", cv2.IMREAD_COLOR)
        # right = cv2.imread("datasets/drivingstereo/2018-07-11-14-48-52/2018-07-11-14-48-52_2018-07-11-14-50-10-570_right.jpg", cv2.IMREAD_COLOR)

        # left = cv2.imread("datasets/eth3d_lowres/forest_2s/im0.png", cv2.IMREAD_COLOR)
        # right = cv2.imread("datasets/eth3d_lowres/forest_2s/im1.png", cv2.IMREAD_COLOR)

        # left = cv2.imread("datasets/eth3d_lowres/playground_3l/im0.png", cv2.IMREAD_COLOR)
        # right = cv2.imread("datasets/eth3d_lowres/playground_3l/im1.png", cv2.IMREAD_COLOR)

        left = cv2.imread("datasets/sceneflow/driving_left.png", cv2.IMREAD_COLOR)
        right = cv2.imread("datasets/sceneflow/driving_right.png", cv2.IMREAD_COLOR)

        # left = cv2.resize (left, (1280,720), cv2.INTER_AREA)
        # right = cv2.resize (right, (1280,720), cv2.INTER_AREA)
        
        imagenet_stats = {'mean': [0.485, 0.456, 0.406], 'std': [0.229, 0.224, 0.225]}
        img_to_tensor_transforms = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(**imagenet_stats),
        ])

        left = img_to_tensor_transforms(left)
        right = img_to_tensor_transforms(right)

        # pad to width and hight to 16 times
        if left.shape[1] % 16 != 0:
            times = left.shape[1]//16       
            top_pad = (times+1)*16 -left.shape[1]
        else:
            top_pad = 0

        if left.shape[2] % 16 != 0:
            times = left.shape[2]//16                       
            right_pad = (times+1)*16-left.shape[2]
        else:
            right_pad = 0    

        left = F.pad(left,(0,right_pad, top_pad,0)).unsqueeze(0)
        right = F.pad(right,(0,right_pad, top_pad,0)).unsqueeze(0)

        output = net (left, right)
        output = output[0].detach().numpy().transpose(1,2,0)
        show_color_disparity ("disparity", output)
        cv2.waitKey(0)

    # save_torchscript(net, "chang-realtime-stereo.scripted.pt", torch.device('cpu'))
    # save_torchscript(net, "chang-realtime-stereo-gpu.scripted.pt", torch.device('cuda'))
    
    # Only tracing worked without substantial changes to the codebase.
    device = torch.device('cpu')
    for w,h in [(1280, 720), (640,480), (320,240), (160,128)]:
        sample_input = (torch.zeros(1,3,h,w).to(device), torch.zeros(1,3,h,w).to(device))
        with torch.no_grad():
            scripted_module = torch.jit.trace(net, sample_input)
            torch.jit.save(scripted_module, f"chang-realtime-stereo-cpu-{w}x{h}.scripted.pt")

        # Need opset16 for grid sampling, currently needs pytorch nightly to do it (1.11 won't).
        # However the exported onnx fails to run:
        # [ONNXRuntimeError] : 1 : FAIL : Load model from chang-realtime-stereo-cpu-320x240.onnx
        #  failed:Type Error: Type parameter (T) of Optype (Mul) bound to different types
        #  (tensor(float) and tensor(int64) in node (Mul_1675).
        if False:
            torch.onnx.export(scripted_module,                 # model being run
                        sample_input,              # model input (or a tuple for multiple inputs)
                        f"chang-realtime-stereo-cpu-{w}x{h}.onnx", # where to save the model (can be a file or file-like object)
                        export_params=True,        # store the trained parameter weights inside the model file
                        opset_version=16,          # the ONNX version to export the model to
                        do_constant_folding=True,  # whether to execute constant folding for optimization
                        input_names = ['left', 'right'],   # the model's input names
                        output_names = ['disparity']) # the model's output names
                        # No dynamic axes with tracing :-(
                        # dynamic_axes={'left' : {0 : 'batch_size', 2 : 'width', 3 : 'height' },  # variable length axes, except channels
                        #             'right' : {0 : 'batch_size', 2 : 'width', 3 : 'height' },
                        #             'output' : {0 : 'batch_size', 2 : 'width', 3 : 'height'}})
    # save_onnx(scripted_module, "chang-realtime-stereo-cpu.onnx")

def benchmark_model(name, size, device):
    w, h = size
    is_gpu = (device == 'cuda')

    def load_model ():
        checkpoint_file = rtstereo_model
        net = RTStereoNet(maxdisp=192, device=device)
        checkpoint = torch.load(checkpoint_file)
        net.load_state_dict(checkpoint['state_dict'])
        net.eval()
        net = net.to (device)
        return net

    def do_inference (model):
        with torch.no_grad():
            sample_input = (torch.zeros(1, 3, h, w).to(device), torch.zeros(1, 3, h, w).to(device))
            outputs = model (*sample_input)
            print (type(outputs))
    
    monitor = StereodemoPerformanceMonitor(f'{name}_{device}_{w}x{h}', load_model, do_inference, is_gpu)
    monitor.run ()
  

if __name__ == "__main__":
    export_models ()
    
    # torch.set_num_threads(1)
    # for s in [(320,240), (640,480), (1280,720)]:
    #     benchmark_model ('chang', s, 'cpu')

    


import cv2
import depthai as dai
import numpy as np
import json

# Adapted from LearnOpenCV
# https://github.com/spmallick/learnopencv/tree/master/oak-getting-started

def getFrame(queue):
  # Get frame from queue
  frame = queue.get()
  # Convert frame to OpenCV format and return
  return frame.getCvFrame()

def getMonoCamera(pipeline, isLeft):
  # Configure mono camera
  mono = pipeline.createMonoCamera()

  # Set Camera Resolution
  mono.setResolution(dai.MonoCameraProperties.SensorResolution.THE_480_P)

  if isLeft:
      # Get left camera
      mono.setBoardSocket(dai.CameraBoardSocket.LEFT)
  else :
      # Get right camera
      mono.setBoardSocket(dai.CameraBoardSocket.RIGHT)
  return mono


def getStereoPair(pipeline, monoLeft, monoRight):
    # Configure stereo pair for depth estimation
    stereo = pipeline.createStereoDepth()
    # Checks occluded pixels and marks them as invalid
    stereo.setLeftRightCheck(True)
    
    # Configure left and right cameras to work as a stereo pair
    monoLeft.out.link(stereo.left)
    monoRight.out.link(stereo.right)

    return stereo

if __name__ == '__main__':

    # Start defining a pipeline
    pipeline = dai.Pipeline()

    # Set up left and right cameras
    monoLeft = getMonoCamera(pipeline, isLeft = True)
    monoRight = getMonoCamera(pipeline, isLeft = False)

    # Combine left and right cameras to form a stereo pair
    stereo = getStereoPair(pipeline, monoLeft, monoRight)
    
    # Set XlinkOut for disparity, rectifiedLeft, and rectifiedRight
    xoutDisp = pipeline.createXLinkOut()
    xoutDisp.setStreamName("disparity")
    
    xoutDepth = pipeline.create(dai.node.XLinkOut)
    xoutDepth.setStreamName("depth")

    xoutRectifiedLeft = pipeline.createXLinkOut()
    xoutRectifiedLeft.setStreamName("rectifiedLeft")

    xoutRectifiedRight = pipeline.createXLinkOut()
    xoutRectifiedRight.setStreamName("rectifiedRight")

    stereo.disparity.link(xoutDisp.input)
    
    stereo.rectifiedLeft.link(xoutRectifiedLeft.input)
    stereo.rectifiedRight.link(xoutRectifiedRight.input)
    stereo.depth.link(xoutDepth.input)
    
    # Pipeline is defined, now we can connect to the device

    with dai.Device(pipeline) as device:

        calib = device.readCalibration()
        w, h = monoLeft.getResolutionSize()
        
        # The rectified stereo images intrinsics both correspond to the right camera intrinsics.
        disparityIntrinsics = calib.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, dai.Size2f(w, h))
        baselineMeters = 1e-2 * calib.getBaselineDistance(dai.CameraBoardSocket.LEFT, dai.CameraBoardSocket.RIGHT)
        with open('stereodemo-calibration.json', 'w') as f:
            d = dict(baseline_meters=baselineMeters,
                     fx=disparityIntrinsics[0][0],
                     fy=disparityIntrinsics[1][1],
                     cx0=disparityIntrinsics[0][2],
                     cx1=disparityIntrinsics[0][2],
                     cy=disparityIntrinsics[1][2])
            f.write(json.dumps(d))

        # Output queues will be used to get the rgb frames and nn data from the outputs defined above
        disparityQueue = device.getOutputQueue(name="disparity", maxSize=1, blocking=False)
        rectifiedLeftQueue = device.getOutputQueue(name="rectifiedLeft", maxSize=1, blocking=False)
        rectifiedRightQueue = device.getOutputQueue(name="rectifiedRight", maxSize=1, blocking=False)
        depthQueue = device.getOutputQueue(name="depth", maxSize=1, blocking=False)

        # Calculate a multiplier for colormapping disparity map
        disparityMultiplier = 255 / stereo.getMaxDisparity()

        cv2.namedWindow("Stereo Pair")
        
        # Variable use to toggle between side by side view and one frame view.
        sideBySide = False

        print ("Press 's' to save an image")

        save_frame_id = 0
        while True:
            
            # Get disparity map
            disparityPixels = getFrame(disparityQueue)
            
            # Colormap disparity for display
            disparity = (disparityPixels * disparityMultiplier).astype(np.uint8)
            disparity = cv2.applyColorMap(disparity, cv2.COLORMAP_JET)
            
            # Depth
            depthMm = getFrame(depthQueue)
            centralDepthMm = depthMm[h//2, w//2]
            centralDisp = disparityPixels[h//2, w//2]
            depthFromDispMeters = (baselineMeters * disparityIntrinsics[0][0]) / centralDisp
            # print (f"Central pixel depth = {centralDepthMm} disparity_raw = {centralDisp} depthFromDispMeters={depthFromDispMeters}")

            # Get left and right rectified frame
            leftFrame = getFrame(rectifiedLeftQueue)
            rightFrame = getFrame(rectifiedRightQueue)
            
            if sideBySide:
                # Show side by side view
                imOut = np.hstack((leftFrame, rightFrame))
            else :
                # Show overlapping frames
                imOut = np.uint8(leftFrame/2 + rightFrame/2)
                        
            imOut = cv2.cvtColor(imOut,cv2.COLOR_GRAY2RGB)
            cv2.imshow("Stereo Pair", imOut)
            cv2.imshow("Disparity", disparity)

            # Check for keyboard input
            key = cv2.waitKey(1)
            if key == ord('q'):
                # Quit when q is pressed
                break
            elif key == ord('t'):
                # Toggle display when t is pressed
                sideBySide = not sideBySide
            elif key == ord('s'):
                # Save the current frames  
                cv2.imwrite(f"img_{save_frame_id:03d}_left.png", leftFrame)
                cv2.imwrite(f"img_{save_frame_id:03d}_right.png", rightFrame)
                print (f"Wrote img_{save_frame_id:03d}_left/right.png")
                save_frame_id += 1




from pathlib import Path
import shutil
import time
from dataclasses import dataclass
import urllib.request
import tempfile
import sys

import torch
from torchvision import transforms

import cv2
import numpy as np

from .methods import Config, EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "chang-realtime-stereo-cpu-1280x720.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-chang-realtimestereo/chang-realtime-stereo-cpu-1280x720.scripted.pt",
    "chang-realtime-stereo-cpu-160x128.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-chang-realtimestereo/chang-realtime-stereo-cpu-160x128.scripted.pt",
    "chang-realtime-stereo-cpu-320x240.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-chang-realtimestereo/chang-realtime-stereo-cpu-320x240.scripted.pt",
    "chang-realtime-stereo-cpu-640x480.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-chang-realtimestereo/chang-realtime-stereo-cpu-640x480.scripted.pt",
}

# https://github.com/JiaRenChang/RealtimeStereo
# I exported the pytorch implementation to torch script via tracing with
# some minor changes to the code https://github.com/JiaRenChang/RealtimeStereo/pull/15
# See chang_realtimestereo_to_torchscript_onnx.py
class ChangRealtimeStereo(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("Chang Real-time (ACCV 2020)",
                         "Attention-Aware Feature Aggregation for Real-time Stereo Matching on Edge Devices. Pre-trained on SceneFlow + Kitti 2015.",
                         {},
                         config)
        self.reset_defaults()

        self.net = None
        self._loaded_model_path = None

        imagenet_stats = {'mean': [0.485, 0.456, 0.406], 'std': [0.229, 0.224, 0.225]}
        self.img_to_tensor_transforms = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(**imagenet_stats),
        ])

    def reset_defaults(self):
        self.parameters.update ({
            "Shape": EnumParameter("Processed image size", 2, ["160x128", "320x240", "640x480", "1280x720"])
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        cols, rows = self.parameters["Shape"].value.split('x')
        cols, rows = int(cols), int(rows)
        self.target_size = (cols, rows)

        model_path = self.config.models_path / f'chang-realtime-stereo-cpu-{cols}x{rows}.scripted.pt'
        self._load_model (model_path)

        left_tensor = self._preprocess_input(input.left_image)
        right_tensor = self._preprocess_input(input.right_image)

        start = time.time()
        with torch.no_grad():
            outputs = self.net(left_tensor, right_tensor)
        elapsed_time = time.time() - start

        disparity_map = self._process_output(outputs)
        if disparity_map.shape[:2] != input.left_image.shape[:2]:
            disparity_map = cv2.resize (disparity_map, (input.left_image.shape[1], input.left_image.shape[0]), cv2.INTER_NEAREST)
            x_scale = input.left_image.shape[1] / float(cols)
            disparity_map *= np.float32(x_scale)

        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _preprocess_input (self, img: np.ndarray):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, self.target_size, cv2.INTER_AREA)
        return self.img_to_tensor_transforms (img).unsqueeze(0)

    def _process_output(self, outputs):
        disparity_map = outputs[0].detach().cpu().permute(1,2,0).numpy()
        return disparity_map

    def _load_model(self, model_path: Path):
        if (self._loaded_model_path == model_path):
            return
        
        if not model_path.exists():
            utils.download_model (urls[model_path.name], model_path)

        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self.net = torch.jit.load(model_path)
        self.net.cpu ()
        self.net.eval ()



from pathlib import Path
import shutil
import time
from dataclasses import dataclass
import urllib.request
import gc
import tempfile
import re
import sys

import torch
from torchvision import transforms

import cv2
import numpy as np

from .methods import Config, EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "sttr-kitti-cpu-240x320-ds1.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cpu-240x320-ds1.scripted.pt",
    "sttr-kitti-cpu-480x640-ds2.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cpu-480x640-ds2.scripted.pt",
    "sttr-kitti-cpu-480x640-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cpu-480x640-ds3.scripted.pt",
    "sttr-kitti-cpu-720x1280-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cpu-720x1280-ds3.scripted.pt",
    "sttr-kitti-cuda-240x320-ds1.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cuda-240x320-ds1.scripted.pt",
    "sttr-kitti-cuda-480x640-ds2.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cuda-480x640-ds2.scripted.pt",
    "sttr-kitti-cuda-480x640-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cuda-480x640-ds3.scripted.pt",
    "sttr-kitti-cuda-720x1280-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-kitti-cuda-720x1280-ds3.scripted.pt",

    "sttr-sceneflow-cpu-240x320-ds1.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cpu-240x320-ds1.scripted.pt",
    "sttr-sceneflow-cpu-480x640-ds2.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cpu-480x640-ds2.scripted.pt",
    "sttr-sceneflow-cpu-480x640-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cpu-480x640-ds3.scripted.pt",
    "sttr-sceneflow-cpu-720x1280-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cpu-720x1280-ds3.scripted.pt",
    "sttr-sceneflow-cuda-240x320-ds1.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cuda-240x320-ds1.scripted.pt",
    "sttr-sceneflow-cuda-480x640-ds2.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cuda-480x640-ds2.scripted.pt",
    "sttr-sceneflow-cuda-480x640-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cuda-480x640-ds3.scripted.pt",
    "sttr-sceneflow-cuda-720x1280-ds3.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-sttr/sttr-sceneflow-cuda-720x1280-ds3.scripted.pt",
}

def clear_gpu_memory():
    gc.collect()
    torch.cuda.empty_cache()    

# https://github.com/mli0603/stereo-transformer
# Made some changes to allow torchscript tracing:
# https://github.com/nburrus/stereo-transformer/commit/0006a022c19f0c7c4d7683408531180a863603a5
class StereoTransformers(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("STereo TRansformer (ICCV 2021)",
                         "Revisiting Stereo Depth Estimation From a Sequence-to-Sequence Perspective with Transformers.",
                         {},
                         config)
        self.reset_defaults()

        self.net = None
        self._loaded_model_path = None
        
        imagenet_stats = {'mean': [0.485, 0.456, 0.406], 'std': [0.229, 0.224, 0.225]}
        self.img_to_tensor_transforms = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(**imagenet_stats),
        ])

    def reset_defaults(self):
        self.parameters.update ({
            "Shape": EnumParameter("Processed image size", 2, ["320x240 (ds1)", "640x480 (ds2)", "640x480 (ds3)", "1280x720 (ds3)"]),
            # "Model": EnumParameter("Pre-trained Model", 0, ["kitti-cpu", "sceneflow-cpu", "kitti-cuda", "sceneflow-cuda"])
            # The CUDA ones segfault with my Python 3.8 venv, but someone worked with Python 3.7.
            # Maybe related to the installed packages instead, need to investigate more.
            # Keeping only the CPU ones for now since it's enough to evaluate.
            "Model": EnumParameter("Pre-trained Model", 0, ["kitti-cpu", "sceneflow-cpu"]),
            "Detect occlusions": EnumParameter("Detect Occlusions", 0, ["Yes", "No"])
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        stereo_output = self._compute_disparity (input)
        clear_gpu_memory ()
        return stereo_output

    def _compute_disparity(self, input: InputPair) -> StereoOutput:
        m = re.match("(\d+)x(\d+) \(ds(\d)\)", self.parameters["Shape"].value)
        cols, rows, downsample = map(lambda v: int(v), m.groups())
        self.target_size = (cols, rows)

        variant = self.parameters["Model"].value
        detect_occlusions = self.parameters["Detect occlusions"].value == "Yes"
        
        model_path = self.config.models_path / f'sttr-{variant}-{rows}x{cols}-ds{downsample}.scripted.pt'
        self._load_model (model_path)

        left_tensor = self._preprocess_input(input.left_image)
        right_tensor = self._preprocess_input(input.right_image)

        col_offset = int(downsample / 2)
        row_offset = int(downsample / 2)
        sampled_cols = torch.arange(col_offset, cols, downsample)[None,]
        sampled_rows = torch.arange(row_offset, rows, downsample)[None,]

        device = torch.device('cuda') if 'cuda' in variant else 'cpu'
        net = self.net.to(device)
        left_tensor = left_tensor.to(device)
        right_tensor = right_tensor.to(device)
        sampled_cols = sampled_cols.to(device)
        sampled_rows = sampled_rows.to(device)

        start = time.time()
        with torch.no_grad():
            outputs = net(left_tensor, right_tensor, sampled_cols, sampled_rows)
        elapsed_time = time.time() - start

        disparity_map = self._process_output(outputs, use_occlusion=detect_occlusions)
        if disparity_map.shape[:2] != input.left_image.shape[:2]:
            disparity_map = cv2.resize (disparity_map, (input.left_image.shape[1], input.left_image.shape[0]), cv2.INTER_NEAREST)
            x_scale = input.left_image.shape[1] / float(cols)
            disparity_map *= np.float32(x_scale)

        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _preprocess_input (self, img: np.ndarray):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, self.target_size, cv2.INTER_AREA)
        return self.img_to_tensor_transforms (img).unsqueeze(0)

    def _process_output(self, outputs, use_occlusion: bool):
        disparity_map = outputs[0][0].detach().cpu().numpy()
        if use_occlusion:
            occ_pred = outputs[1][0].data.cpu().numpy() > 0.5
            disparity_map[occ_pred] = -1.0
        return disparity_map

    def _load_model(self, model_path: Path):
        # FIXME: always reload the model, for some reason
        # feeding multiple images to the same model freezes
        # with CUDA. Maybe due to multi-threading?
        # if (self._loaded_model_path == model_path):
        #     return
        
        if not model_path.exists():
            utils.download_model (urls[model_path.name], model_path)

        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self.net = torch.jit.load(model_path)
        self.net.eval ()



from abc import abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Tuple
import time
import json
import numpy as np


@dataclass
class Calibration:
    width: int
    height: int
    fx: float
    fy: float
    cx0: float  # cx is the only one that can differ between both cameras.
    cx1: float
    cy: float
    baseline_meters: float
    depth_range: Tuple[float] = (0.3, 20.0)
    left_image_rect_normalized: np.ndarray = field(
        default_factory=lambda: np.array([0., 0., 1., 1.])) # origin, size in percent of image size
    comment: str = ""

    def to_json(self):
        return json.dumps(self.__dict__)

    @staticmethod
    def from_json(json_str):
        d = json.loads(json_str)
        return Calibration(**d)

    def downsample(self, new_width: int, new_height: int):
        sx = new_width / self.width
        sy = new_height / self.height
        self.width = new_width
        self.height = new_height
        self.fx *= sx
        self.fy *= sy
        self.cx0 *= sx
        self.cx1 *= sx
        self.cy *= sy


@dataclass
class InputPair:
    left_image: np.ndarray
    right_image: np.ndarray
    calibration: Calibration
    status: str
    input_disparity: np.ndarray = None

    def has_data(self):
        return self.left_image is not None


@dataclass
class StereoOutput:
    disparity_pixels: np.ndarray
    color_image_bgr: np.ndarray
    computation_time: float
    point_cloud: Any = None
    disparity_color: np.ndarray = None


@dataclass
class IntParameter:
    description: str
    value: int
    min: int
    max: int
    to_valid: Any = lambda x: x  # default is to just accept anything

    def set_value(self, x: int):
        self.value = self.to_valid(x)


@dataclass
class EnumParameter:
    description: str
    index: int  # index in the list
    values: List[str]

    def set_index(self, idx: int):
        self.index = idx

    def set_value(self, value):
        self.index = self.values.index(value)

    @property
    def value(self) -> str:
        return self.values[self.index]


@dataclass
class Config:
    models_path: Path


class StereoMethod:
    def __init__(self, name: str, description: str, parameters: Dict, config: Config):
        self.name = name
        self.parameters = parameters
        self.description = description
        self.config = config

    def reset_defaults(self):
        pass

    @abstractmethod
    def compute_disparity(self, input: InputPair) -> StereoOutput:
        """Return the disparity map in pixels and the actual computation time.

        Both input images are assumed to be rectified.
        """
        return StereoOutput(None, None, None, None)

    @staticmethod
    def depth_meters_from_disparity(disparity_pixels: np.ndarray, calibration: Calibration):
        old_seterr = np.seterr(divide='ignore')
        dcx = np.float32(calibration.cx0 - calibration.cx1)
        depth_meters = np.float32(calibration.baseline_meters * calibration.fx) / (disparity_pixels - dcx)
        depth_meters = np.nan_to_num(depth_meters)
        depth_meters[disparity_pixels < 0.] = -1.0
        np.seterr(**old_seterr)
        return depth_meters

    @staticmethod
    def disparity_from_depth_meters(depth_meters: np.ndarray, calibration: Calibration):
        old_seterr = np.seterr(divide='ignore')
        dcx = np.float32(calibration.cx0 - calibration.cx1)
        disparity_pixels = (np.float32(calibration.baseline_meters * calibration.fx) / depth_meters) + dcx
        disparity_pixels = np.nan_to_num(disparity_pixels)
        np.seterr(**old_seterr)
        return disparity_pixels


#!/usr/bin/env python3

import tempfile
import unittest
from pathlib import Path

import cv2
import numpy as np
import math

from stereodemo import method_opencv_bm
from stereodemo import method_chang_realtime_stereo
from stereodemo import method_hitnet
from stereodemo import method_cre_stereo
from stereodemo import method_raft_stereo
from stereodemo import method_sttr
from stereodemo.methods import Config, InputPair, Calibration, StereoOutput, StereoMethod

data_folder = Path(__file__).parent.parent / 'datasets' / 'eth3d_lowres' / 'delivery_area_1l'
left_image = cv2.imread (str(data_folder / 'im0.png'), cv2.IMREAD_COLOR)
right_image = cv2.imread (str(data_folder / 'im1.png'), cv2.IMREAD_COLOR)
calib_file = data_folder / 'stereodemo_calibration.json'
calibration = Calibration.from_json (open(calib_file, 'r').read())
input = InputPair (left_image, right_image, calibration, "Test images loaded.", None)
models_path = Path(tempfile.gettempdir()) / 'models'
models_path.mkdir(parents=True, exist_ok=True)
config = Config(models_path)

class TestStereoInference(unittest.TestCase):

    def check_method(self, method: StereoMethod, expected_median: float, expected_coverage: float):
        output = method.compute_disparity (input)
        valid_pixels = output.disparity_pixels[output.disparity_pixels > 0.]
        coverage = valid_pixels.size / output.disparity_pixels.size
        median_value =  np.median (valid_pixels)
        self.assertAlmostEqual (median_value, expected_median, delta=0.01)
        self.assertAlmostEqual (coverage, expected_coverage, delta=0.01)

    def test_bm(self):
        self.check_method (method_opencv_bm.StereoBM(config), 4.8125, 0.4403)

    def test_sgbm(self):
        self.check_method (method_opencv_bm.StereoSGBM(config), 5.1875, 0.8515)

    def test_chang_realtime(self):
        m = method_chang_realtime_stereo.ChangRealtimeStereo(config)
        m.parameters["Shape"].set_value ("320x240")
        self.check_method (m, 12.7776, 1.0)

    def test_hitnet(self):
        m = method_hitnet.HitnetStereo(config)
        m.parameters["Shape"].set_value ("320x240")
        self.check_method (m, 4.9103, 1.0)

    def test_crestereo(self):
        m = method_cre_stereo.CREStereo(config)
        m.parameters["Shape"].set_value ("320x240")
        self.check_method (m, 4.6287, 1.0)

    def test_raft_stereo(self):
        m = method_raft_stereo.RaftStereo(config)
        m.parameters["Shape"].set_value ("320x256")
        self.check_method (m, 4.6408, 1.0)

    def test_sttr(self):
        m = method_sttr.StereoTransformers(config)
        m.parameters["Shape"].set_value ("640x480 (ds3)")
        self.check_method (m, 7.4636, 0.9869)

if __name__ == '__main__':
    unittest.main()



from pathlib import Path
import time
from dataclasses import dataclass

import cv2
import numpy as np
import onnxruntime

from .methods import Config, EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "crestereo_combined_iter10_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter10_240x320.onnx",
    "crestereo_combined_iter10_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter10_480x640.onnx",
    "crestereo_combined_iter10_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter10_720x1280.onnx",
    "crestereo_combined_iter20_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter20_240x320.onnx",
    "crestereo_combined_iter20_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter20_480x640.onnx",
    "crestereo_combined_iter20_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter20_720x1280.onnx",
    "crestereo_combined_iter2_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter2_240x320.onnx",
    "crestereo_combined_iter2_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter2_480x640.onnx",
    "crestereo_combined_iter2_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter2_720x1280.onnx",
    "crestereo_combined_iter5_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter5_240x320.onnx",
    "crestereo_combined_iter5_380x480.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter5_380x480.onnx",
    "crestereo_combined_iter5_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter5_480x640.onnx",
    "crestereo_combined_iter5_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_combined_iter5_720x1280.onnx",
    "crestereo_init_iter10_180x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter10_180x320.onnx",
    "crestereo_init_iter10_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter10_240x320.onnx",
    "crestereo_init_iter10_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter10_480x640.onnx",
    "crestereo_init_iter10_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter10_720x1280.onnx",
    "crestereo_init_iter20_180x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter20_180x320.onnx",
    "crestereo_init_iter20_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter20_240x320.onnx",
    "crestereo_init_iter20_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter20_480x640.onnx",
    "crestereo_init_iter20_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter20_720x1280.onnx",
    "crestereo_init_iter2_180x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter2_180x320.onnx",
    "crestereo_init_iter2_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter2_240x320.onnx",
    "crestereo_init_iter2_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter2_480x640.onnx",
    "crestereo_init_iter2_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter2_720x1280.onnx",
    "crestereo_init_iter5_180x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter5_180x320.onnx",
    "crestereo_init_iter5_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter5_240x320.onnx",
    "crestereo_init_iter5_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter5_480x640.onnx",
    "crestereo_init_iter5_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-crestereo/crestereo_init_iter5_720x1280.onnx"
}

# Adapted from https://github.com/ibaiGorordo/ONNX-CREStereo-Depth-Estimation
# https://github.com/PINTO0309/PINTO_model_zoo/tree/main/284_CREStereo
class CREStereo(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("CRE Stereo (CVPR 2022)", 
                         "Practical Stereo Matching via Cascaded Recurrent Network with Adaptive Correlation. Pre-trained on a large range of datasets.", 
                         {},
                         config)
        self.reset_defaults()

        self._loaded_session = None
        self._loaded_model_path = None

    def reset_defaults(self):
        self.parameters.update ({
            "Iterations": EnumParameter("Number of iterations", 1, ["2", "5", "10", "20"]),
            "Mode": EnumParameter("Number of passes. The combined version does 2 passes, one to get an initial estimation and a second one to refine it.",
                                  1, ["init", "combined"]),
            "Shape": EnumParameter("Processed image size", 1, ["320x240", "640x480", "1280x720"])            
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        left_image, right_image = input.left_image, input.right_image
        cols, rows = self.parameters["Shape"].value.split('x')
        version = self.parameters["Mode"].value
        iters = self.parameters["Iterations"].value
        model_path = self.config.models_path / f'crestereo_{version}_iter{iters}_{rows}x{cols}.onnx'
        self._load_model (model_path)

        left_tensor = self._prepare_input(left_image)
        right_tensor = self._prepare_input(right_image)

        # Get the half resolution to calculate flow_init
        if self._has_flow:
            left_tensor_half = self._prepare_input(left_image, half=True)
            right_tensor_half = self._prepare_input(right_image, half=True)
            start = time.time()
            outputs = self._inference_with_flow(left_tensor_half,
                                               right_tensor_half,
                                               left_tensor,
                                               right_tensor)
        else:
            # Estimate the disparity map
            start = time.time()
            outputs = self._inference_without_flow(left_tensor, right_tensor)

        elapsed_time = time.time() - start
        
        disparity_map = self.process_output(outputs)

        if disparity_map.shape[:2] != left_image.shape[:2]:
            disparity_map = cv2.resize (disparity_map, (left_image.shape[1], left_image.shape[0]), cv2.INTER_NEAREST)
            x_scale = left_image.shape[1] / float(cols)
            disparity_map *= np.float32(x_scale)        
        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _download_model (self, model_path: Path):
        utils.download_model (urls[model_path.name], model_path)

    def _load_model(self, model_path: Path):
        if (self._loaded_model_path == model_path):
            return
        
        if not model_path.exists():
            self._download_model (model_path)

        # To try with just one CPU core.
        # opts = onnxruntime.SessionOptions()
        # opts.intra_op_num_threads = 1
        # opts.inter_op_num_threads = 1
        # opts.execution_mode = onnxruntime.ExecutionMode.ORT_SEQUENTIAL
        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self._loaded_session = onnxruntime.InferenceSession(str(model_path), providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])
        
        # Get model info
        self.load_input_details()
        self.load_output_details()

        # Check if the model has init flow
        self._has_flow = len(self.input_names) > 2

    def _prepare_input(self, img, half=False):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        if half:
            img_input = cv2.resize(
                img, (self.input_width//2, self.input_height//2), cv2.INTER_AREA)
        else:
            img_input = cv2.resize(
                img, (self.input_width, self.input_height), cv2.INTER_AREA)
        img_input = img_input.transpose(2, 0, 1)
        img_input = img_input[np.newaxis, :, :, :]
        return img_input.astype(np.float32)

    def _inference_without_flow(self, left_tensor, right_tensor):

        return self._loaded_session.run(self.output_names, {self.input_names[0]: left_tensor,
                                                            self.input_names[1]: right_tensor})[0]

    def _inference_with_flow(self, left_tensor_half, right_tensor_half, left_tensor, right_tensor):

        return self._loaded_session.run(self.output_names, {self.input_names[0]: left_tensor_half,
                                                            self.input_names[1]: right_tensor_half,
                                                            self.input_names[2]: left_tensor,
                                                            self.input_names[3]: right_tensor})[0]

    def process_output(self, output):
        return np.squeeze(output[:, 0, :, :])

    def load_input_details(self):
        model_inputs = self._loaded_session.get_inputs()
        self.input_names = [
            model_inputs[i].name for i in range(len(model_inputs))]

        self.input_shape = model_inputs[-1].shape
        self.input_height = self.input_shape[2]
        self.input_width = self.input_shape[3]

    def load_output_details(self):
        model_outputs = self._loaded_session.get_outputs()
        self.output_names = [
            model_outputs[i].name for i in range(len(model_outputs))]

        self.output_shape = model_outputs[0].shape



from abc import abstractmethod
import json
from pathlib import Path
import sys
import tempfile
import time
from types import SimpleNamespace
from typing import List

import numpy as np

import cv2


from . import visualizer
from . import methods

from .method_opencv_bm import StereoBM, StereoSGBM
from .method_raft_stereo import RaftStereo
from .method_cre_stereo import CREStereo
from .method_chang_realtime_stereo import ChangRealtimeStereo
from .method_hitnet import HitnetStereo
from .method_sttr import StereoTransformers
from stereodemo.method_dist_depth import DistDepth

def parse_args():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--oak', action='store_true', help='Use an oak-D camera to grab images.')
    parser.add_argument('--oak-output-folder', type=Path, default=None, help='Output folder to save the images grabbed by the OAK camera')
    parser.add_argument('images',
                        help='rectified_left1 rectified_right1 ... [rectified_leftN rectified_rightN]. Load image pairs from disk. You can also specify folders.',
                        type=Path, 
                        default=None,
                        nargs='*')
    parser.add_argument('--calibration', type=Path, help='Calibration json. If unspecified, it will try to load a stereodemo_calibration.json file in the left image parent folder.', default=None)
    default_models_path = Path.home() / ".cache" / "stereodemo" / "models"
    parser.add_argument('--models-path', type=Path, help='Path to store the downloaded models.', default=default_models_path)
    return parser.parse_args()

def find_stereo_images_in_dir(dir: Path):
    left_files = []
    right_files = []

    def validated_lists():
        for f in right_files:
            assert f.exists()
        return left_files, right_files

    for ext in ['jpg', 'png']:
        left = sorted(list(dir.glob(f'**/*left*.{ext}')))
        if len(left) != 0:
            right = [f.parent / f.name.replace('left', 'right') for f in left]
            left_files += left
            right_files += right

    for ext in ['jpg', 'png']:
        left = sorted(list(dir.glob(f'**/im0.{ext}')))
        if len(left) != 0:
            right = [f.parent / f.name.replace('im0', 'im1') for f in left]
            left_files += left
            right_files += right            
    
    return validated_lists()

class FileListSource (visualizer.Source):
    def __init__(self, file_or_dir_list, calibration=None):        
        self.left_images_path = []
        self.right_images_path = []

        while file_or_dir_list:
            f = file_or_dir_list.pop(0)
            if f.is_dir():
                left, right = find_stereo_images_in_dir (f)
                self.left_images_path += left
                self.right_images_path += right
            else:
                if f.suffix.lower() not in ['.png', '.jpg', '.jpeg']:
                    print (f"Warning: ignoring {f}, not an image extension.")
                    continue
                try:
                    right_f = file_or_dir_list.pop(0)
                except:
                    print (f"Missing right image for {f}, skipping")
                    continue
                self.left_images_path.append(f)
                self.right_images_path.append(right_f)
        
        self.index = 0
        self.user_provided_calibration_path = calibration
        self.num_pairs = len(self.left_images_path)
        if self.num_pairs == 0:
            raise Exception("No image pairs.")

    def is_live(self):
        return False

    def selected_index (self) -> int:
        return self.index

    def get_pair_at_index(self, idx: int) -> methods.InputPair:
        self.index = idx
        
        def load_image(path):
            im =  cv2.imread(str(path), cv2.IMREAD_COLOR)
            assert im is not None
            return im

        left_image_path = self.left_images_path[self.index]
        left_image = load_image(left_image_path)
        if self.user_provided_calibration_path is None:
            calibration_path = left_image_path.parent / 'stereodemo_calibration.json'
            if not calibration_path.exists():
                print (f"Warning: no calibration file found {calibration_path}. Using default calibration, the point cloud won't be accurate.")
                calibration_path = None
        else:
            calibration_path = self.user_provided_calibration_path
        if calibration_path:
            calib = visualizer.Calibration.from_json (open(calibration_path, 'r').read())
        else:
            # Fake reasonable calibration.
            calib = visualizer.Calibration(left_image.shape[1],
                                           left_image.shape[0],
                                           left_image.shape[0]*0.8,
                                           left_image.shape[0]*0.8,
                                           left_image.shape[1]/2.0, # cx0
                                           left_image.shape[1]/2.0, # cx1
                                           left_image.shape[0]/2.0,
                                           0.075)
            
        right_image_path = self.right_images_path[self.index]
        status = f"{left_image_path} / {right_image_path}"
        return visualizer.InputPair (left_image, load_image(right_image_path), calib, status)

    def get_pair_list(self) -> List[str]:
        return [str(f) for f in self.left_images_path]

    def get_next_pair(self):        
        self.index = (self.index + 1) % self.num_pairs
        return self.get_pair_at_index(self.index)

def main():
    args = parse_args()

    try:
        args.models_path.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        sys.stderr.write (f"Warning: cannot use the default models path {args.models_path}: {e}\n")
        sys.stderr.write ("A valid path is necessary to store the downloaded models.\n")
        args.models_path = Path(tempfile.gettempdir()) / 'stereodemo_models'
        sys.stderr.write (f"Going to use the temporary directory {args.models_path} instead, specify --model-paths to specify a custom persistent path instead.\n")
        try:
            args.models_path.mkdir(parents=True, exist_ok=True)
        except:
            sys.stderr.write ("Could not create a temporary directory to store the downloaded models.\n")
            sys.stderr.write ("Aborting, you need to specify --models-path with a valid writable path.\n")
            sys.exit (1)
    print (f"INFO: will store downloaded models in {args.models_path}")

    config = methods.Config(args.models_path)
    method_list = [
        StereoBM(config),
        StereoSGBM(config),
        CREStereo(config),
        RaftStereo(config),
        HitnetStereo(config),
        StereoTransformers(config),
        ChangRealtimeStereo(config),
        DistDepth(config)
    ]

    if args.images:
        source = FileListSource(args.images, args.calibration)
    elif args.oak:
        from .oakd_source import OakdSource, StereoFromOakInputSource
        source = OakdSource(args.oak_output_folder)
        method_list = [StereoFromOakInputSource(config)] + method_list
    else:
        datasets_path = Path(__file__).parent / 'datasets'
        if not datasets_path.exists():
            print (f"Tried but failed to find files in {datasets_path}")
            print ("You need to specify --oak or provide images")
            sys.exit (1)
        source = FileListSource([datasets_path], args.calibration)

    method_dict = { method.name:method for method in method_list } 

    viz = visualizer.Visualizer(method_dict, source)

    while True:
        start_time = time.time()
        if not viz.update_once ():
            break
        cv2.waitKey (1)
        elapsed = time.time() - start_time
        time_to_sleep = 1/30.0 - elapsed
        if time_to_sleep > 0:
            time.sleep (time_to_sleep)





from pathlib import Path

import numpy as np

import tempfile
import urllib.request
import shutil
import sys

def download_model (url: str, model_path: Path):
    filename = model_path.name
    with tempfile.TemporaryDirectory() as d:
        tmp_file_path = Path(d) / filename
        print (f"Downloading {filename} from {url} to {model_path}...")
        urllib.request.urlretrieve(url, tmp_file_path)
        shutil.move (tmp_file_path, model_path)

def pad_width (size: int, multiple: int):
    return 0 if size % multiple == 0 else multiple - (size%multiple)

class ImagePadder:
    def __init__(self, multiple, mode):
        self.multiple = multiple
        self.mode = mode
    
    def pad (self, im: np.ndarray):
        # H,W,C
        rows = im.shape[0]
        cols = im.shape[1]
        self.rows_to_pad = pad_width(rows, self.multiple)
        self.cols_to_pad = pad_width(cols, self.multiple)
        if self.rows_to_pad == 0 and self.cols_to_pad == 0:
            return im
        return np.pad (im, ((0, self.rows_to_pad), (0, self.cols_to_pad), (0, 0)), mode=self.mode)

    def unpad (self, im: np.ndarray):        
        w = im.shape[1] - self.cols_to_pad
        h = im.shape[0] - self.rows_to_pad
        return im[:h, :w, :]



from pathlib import Path
import shutil
import time
from dataclasses import dataclass
import urllib.request
import gc
import tempfile
import sys

import torch
from torchvision import transforms

import cv2
import numpy as np

from .methods import Config, EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "raft-stereo-eth3d-cpu-128x160.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cpu-128x160.scripted.pt",
    "raft-stereo-eth3d-cpu-256x320.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cpu-256x320.scripted.pt",
    "raft-stereo-eth3d-cpu-480x640.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cpu-480x640.scripted.pt",
    "raft-stereo-eth3d-cpu-736x1280.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cpu-736x1280.scripted.pt",
    "raft-stereo-eth3d-cuda-128x160.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cuda-128x160.scripted.pt",
    "raft-stereo-eth3d-cuda-256x320.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cuda-256x320.scripted.pt",
    "raft-stereo-eth3d-cuda-480x640.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cuda-480x640.scripted.pt",
    "raft-stereo-eth3d-cuda-736x1280.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-eth3d-cuda-736x1280.scripted.pt",
    "raft-stereo-fast-cpu-128x160.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cpu-128x160.scripted.pt",
    "raft-stereo-fast-cpu-256x320.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cpu-256x320.scripted.pt",
    "raft-stereo-fast-cpu-480x640.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cpu-480x640.scripted.pt",
    "raft-stereo-fast-cpu-736x1280.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cpu-736x1280.scripted.pt",
    "raft-stereo-fast-cuda-128x160.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cuda-128x160.scripted.pt",
    "raft-stereo-fast-cuda-256x320.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cuda-256x320.scripted.pt",
    "raft-stereo-fast-cuda-480x640.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cuda-480x640.scripted.pt",
    "raft-stereo-fast-cuda-736x1280.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-fast-cuda-736x1280.scripted.pt",
    "raft-stereo-middlebury-cpu-128x160.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cpu-128x160.scripted.pt",
    "raft-stereo-middlebury-cpu-256x320.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cpu-256x320.scripted.pt",
    "raft-stereo-middlebury-cpu-480x640.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cpu-480x640.scripted.pt",
    "raft-stereo-middlebury-cpu-736x1280.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cpu-736x1280.scripted.pt",
    "raft-stereo-middlebury-cuda-128x160.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cuda-128x160.scripted.pt",
    "raft-stereo-middlebury-cuda-256x320.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cuda-256x320.scripted.pt",
    "raft-stereo-middlebury-cuda-480x640.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cuda-480x640.scripted.pt",
    "raft-stereo-middlebury-cuda-736x1280.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-raft-stereo/raft-stereo-middlebury-cuda-736x1280.scripted.pt",
}

def clear_gpu_memory():
    gc.collect()
    torch.cuda.empty_cache()    

# https://github.com/princeton-vl/RAFT-Stereo
# I exported the pytorch implementation to torch script via tracing, with minor modifications of the source code.
# https://github.com/nburrus/RAFT-Stereo/commit/ebbb5a807227927ab4551274039e9bdd16a1b010
# Their fastest implementation was not imported.
class RaftStereo(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("RAFT-Stereo (3DV 2021)",
                         "RAFT-Stereo: Multilevel Recurrent Field Transforms for Stereo Matching.",
                         {},
                         config)
        self.reset_defaults()

        self.net = None
        self._loaded_model_path = None

    def reset_defaults(self):
        self.parameters.update ({
            "Shape": EnumParameter("Processed image size", 2, ["160x128", "320x256", "640x480", "1280x736"]),
            # "Model": EnumParameter("Pre-trained Model", 1, ["eth3d-cuda", "eth3d-cpu", "fast-cuda", "fast-cpu", "middlebury-cuda"])
            # The eth3d and fast cuda models required --corr_implementation alt to work once loaded via torchscript.
            # The supposedly faster "reg" is not working with a torch/cuda segfault, not sure why.
            "Model": EnumParameter("Pre-trained Model", 0, ["fast-cpu", "middlebury-cpu", "eth3d-cpu", "fast-cuda", "middlebury-cuda", "eth3d-cuda"])
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        stereo_output = self._compute_disparity (input)
        clear_gpu_memory ()
        return stereo_output

    def _compute_disparity(self, input: InputPair) -> StereoOutput:
        cols, rows = self.parameters["Shape"].value.split('x')
        cols, rows = int(cols), int(rows)
        self.target_size = (cols, rows)

        variant = self.parameters["Model"].value
        
        model_path = self.config.models_path / f'raft-stereo-{variant}-{rows}x{cols}.scripted.pt'
        self._load_model (model_path)

        left_tensor = self._preprocess_input(input.left_image)
        right_tensor = self._preprocess_input(input.right_image)

        device = torch.device('cuda') if 'cuda' in variant else 'cpu'
        net = self.net.to(device)
        left_tensor = left_tensor.to(device)
        right_tensor = right_tensor.to(device)

        start = time.time()
        with torch.no_grad():
            outputs = self.net(left_tensor, right_tensor)
        elapsed_time = time.time() - start

        disparity_map = self._process_output(outputs)
        if disparity_map.shape[:2] != input.left_image.shape[:2]:
            disparity_map = cv2.resize (disparity_map, (input.left_image.shape[1], input.left_image.shape[0]), cv2.INTER_NEAREST)
            x_scale = input.left_image.shape[1] / float(cols)
            disparity_map *= np.float32(x_scale)

        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _preprocess_input (self, img: np.ndarray):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, self.target_size, cv2.INTER_AREA)
        # -> C,H,W
        # Normalization done in the model itself.
        return torch.from_numpy(img).permute(2, 0, 1).unsqueeze(0).float()

    def _process_output(self, outputs):
        disparity_map = outputs[1][0].detach().cpu().squeeze(0).squeeze(0).numpy() * -1.0
        return disparity_map

    def _load_model(self, model_path: Path):
        # FIXME: always reload the model, for some reason
        # feeding multiple images to the same model freezes
        # with CUDA. Maybe due to multi-threading?
        # if (self._loaded_model_path == model_path):
        #     return
        
        if not model_path.exists():
            utils.download_model (urls[model_path.name], model_path)

        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self.net = torch.jit.load(model_path)
        self.net.eval ()



from pathlib import Path
import shutil
import time
from dataclasses import dataclass
import urllib.request
import tempfile
import sys

import onnxruntime

import cv2
import numpy as np

from .methods import Config, EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "hitnet_eth3d_120x160.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_eth3d_120x160.onnx",
    "hitnet_eth3d_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_eth3d_240x320.onnx",
    "hitnet_eth3d_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_eth3d_480x640.onnx",
    "hitnet_eth3d_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_eth3d_720x1280.onnx",
    "hitnet_middlebury_120x160.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_middlebury_120x160.onnx",
    "hitnet_middlebury_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_middlebury_240x320.onnx",
    "hitnet_middlebury_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_middlebury_480x640.onnx",
    "hitnet_middlebury_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_middlebury_720x1280.onnx",
    "hitnet_sceneflow_120x160.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_sceneflow_120x160.onnx",
    "hitnet_sceneflow_240x320.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_sceneflow_240x320.onnx",
    "hitnet_sceneflow_480x640.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_sceneflow_480x640.onnx",
    "hitnet_sceneflow_720x1280.onnx": "https://github.com/nburrus/stereodemo/releases/download/v0.1-hitnet/hitnet_sceneflow_720x1280.onnx",
}

# Adapted from https://github.com/ibaiGorordo/ONNX-HITNET-Stereo-Depth-estimation
# Onnx models from https://github.com/PINTO0309/PINTO_model_zoo/tree/main/142_HITNET
# Official implementation https://github.com/google-research/google-research/tree/master/hitnet
class HitnetStereo(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("Hitnet (CVPR 2021)",
                         "HITNet: Hierarchical Iterative Tile Refinement Network for Real-time Stereo Matching",
                         {},
                         config)
        self.reset_defaults()

        self._loaded_session = None
        self._loaded_model_path = None

    def reset_defaults(self):
        self.parameters.update ({
            "Shape": EnumParameter("Processed image size", 2, ["160x120", "320x240", "640x480", "1280x720"]),
            "Training Set": EnumParameter("Dataset used during training", 1, ["sceneflow", "middlebury", "eth3d"])
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:        
        cols, rows = self.parameters["Shape"].value.split('x')
        cols, rows = int(cols), int(rows)
        training_set = self.parameters["Training Set"].value

        model_path = self.config.models_path / f'hitnet_{training_set}_{rows}x{cols}.onnx'
        self._load_model (model_path)

        model_inputs = self._loaded_session.get_inputs()
        model_outputs = self._loaded_session.get_outputs()
        model_rows, model_cols = model_inputs[0].shape[2:] # B,C,H,W
        self.target_size = (model_cols, model_rows)

        grayscale = True if training_set == 'eth3d' else False
        combined_tensor = self._preprocess_input(input.left_image, input.right_image, grayscale)

        start = time.time()
        input_names = [model_inputs[i].name for i in range(len(model_inputs))]
        output_names = [model_outputs[i].name for i in range(len(model_outputs))]
        outputs = self._loaded_session.run(['reference_output_disparity'], { 'input': combined_tensor })
        elapsed_time = time.time() - start

        disparity_map = self._process_output(outputs)
        if disparity_map.shape[:2] != input.left_image.shape[:2]:
            model_output_cols = disparity_map.shape[1]
            disparity_map = cv2.resize (disparity_map, (input.left_image.shape[1], input.left_image.shape[0]), cv2.INTER_NEAREST)
            x_scale = input.left_image.shape[1] / float(model_output_cols)
            disparity_map *= np.float32(x_scale)

        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _preprocess_input (self, left: np.ndarray, right: np.ndarray, grayscale: bool):
        if grayscale:
            # H,W
            left = cv2.cvtColor(left, cv2.COLOR_BGR2GRAY)
            right = cv2.cvtColor(right, cv2.COLOR_BGR2GRAY)
        else:
            # H,W,C=3
            left = cv2.cvtColor(left, cv2.COLOR_BGR2RGB)
            right = cv2.cvtColor(right, cv2.COLOR_BGR2RGB)

        left = cv2.resize(left, self.target_size, cv2.INTER_AREA)
        right = cv2.resize(right, self.target_size, cv2.INTER_AREA)

        # Grayscale needs expansion to reach H,W,C.
        # Need to do that now because resize would change the shape.
        if left.ndim == 2:
            left = left[..., np.newaxis]
            right = right[..., np.newaxis]

        # -> H,W,C=2 or 6 , normalized to [0,1]
        combined_img = np.concatenate((left, right), axis=-1) / 255.0
        # -> C,H,W
        combined_img = combined_img.transpose(2, 0, 1)
        # -> B=1,C,H,W
        combined_img = np.expand_dims(combined_img, 0).astype(np.float32)
        return combined_img

    def _process_output(self, outputs):
        disparity_map = outputs[0][0].squeeze(-1)
        return disparity_map

    def _load_model(self, model_path: Path):
        if (self._loaded_model_path == model_path):
            return
        
        if not model_path.exists():
            utils.download_model (urls[model_path.name], model_path)

        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self._loaded_session = onnxruntime.InferenceSession(str(model_path), providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])



from pathlib import Path
import sys
import time

from . import visualizer
from .methods import Config, InputPair, StereoMethod, StereoOutput

import cv2

try:
    import depthai as dai
except ImportError:
    print ("You need to 'pip install depthai' to use the OAK camera.")
    sys.exit (1)

def getFrame(queue):
  # Get frame from queue
  frame = queue.get()
  # Convert frame to OpenCV format and return
  return frame.getCvFrame()

def getMonoCamera(pipeline, isLeft):
  # Configure mono camera
  mono = pipeline.createMonoCamera()

  # Set Camera Resolution
  mono.setResolution(dai.MonoCameraProperties.SensorResolution.THE_480_P)

  if isLeft:
      # Get left camera
      mono.setBoardSocket(dai.CameraBoardSocket.LEFT)
  else :
      # Get right camera
      mono.setBoardSocket(dai.CameraBoardSocket.RIGHT)
  return mono

def getStereoPair(pipeline, monoLeft, monoRight):
    # Configure stereo pair for depth estimation
    stereo = pipeline.createStereoDepth()
    # Checks occluded pixels and marks them as invalid
    stereo.setLeftRightCheck(True)
    
    # Configure left and right cameras to work as a stereo pair
    monoLeft.out.link(stereo.left)
    monoRight.out.link(stereo.right)

    return stereo

class OakdSource (visualizer.Source):
    def __init__(self, output_folder: Path = None):
        self.connect ()
        self.output_folder = output_folder
        self.frameIndex = 0

    def connect (self):
        print ("Trying to connect to an OAK camera...")
        pipeline = dai.Pipeline()

        # Set up left and right cameras
        monoLeft = getMonoCamera(pipeline, isLeft = True)
        monoRight = getMonoCamera(pipeline, isLeft = False)

        # Combine left and right cameras to form a stereo pair
        stereo = getStereoPair(pipeline, monoLeft, monoRight)

        
        # Set XlinkOut for disparity, rectifiedLeft, and rectifiedRight
        xoutDisp = pipeline.createXLinkOut()
        xoutDisp.setStreamName("disparity")
        
        # xoutDepth = pipeline.create(dai.node.XLinkOut)
        # xoutDepth.setStreamName("depth")

        xoutRectifiedLeft = pipeline.createXLinkOut()
        xoutRectifiedLeft.setStreamName("rectifiedLeft")

        xoutRectifiedRight = pipeline.createXLinkOut()
        xoutRectifiedRight.setStreamName("rectifiedRight")

        stereo.disparity.link(xoutDisp.input)
        
        stereo.rectifiedLeft.link(xoutRectifiedLeft.input)
        stereo.rectifiedRight.link(xoutRectifiedRight.input)
        # stereo.depth.link(xoutDepth.input)

        self.device = dai.Device(pipeline).__enter__()

        oak_calib = self.device.readCalibration()
        w, h = monoLeft.getResolutionSize()

        # Intrinsics of disparity are taken from the right image.
        disparityIntrinsics = oak_calib.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, dai.Size2f(w, h))
        baselineMeters = 1e-2 * oak_calib.getBaselineDistance(dai.CameraBoardSocket.LEFT, dai.CameraBoardSocket.RIGHT)

        self.calibration = visualizer.Calibration(w, h,
                                                  fx=disparityIntrinsics[0][0],
                                                  fy=disparityIntrinsics[1][1],
                                                  cx0=disparityIntrinsics[0][2],
                                                  cx1=disparityIntrinsics[0][2],
                                                  cy=disparityIntrinsics[1][2],
                                                  baseline_meters=baselineMeters)

        # Output queues will be used to get the rgb frames and nn data from the outputs defined above
        self.disparityQueue = self.device.getOutputQueue(name="disparity", maxSize=1, blocking=False)
        self.rectifiedLeftQueue = self.device.getOutputQueue(name="rectifiedLeft", maxSize=1, blocking=False)
        self.rectifiedRightQueue = self.device.getOutputQueue(name="rectifiedRight", maxSize=1, blocking=False)
        # depthQueue = self.device.getOutputQueue(name="depth", maxSize=1, blocking=False)

    def is_live(self):
        return True

    def get_next_pair(self):
        leftFrame = getFrame(self.rectifiedLeftQueue)
        rightFrame = getFrame(self.rectifiedRightQueue)
        if self.output_folder is not None:
            self.output_folder.mkdir(parents=True, exist_ok=True)
            cv2.imwrite(str(self.output_folder / f"img_{self.frameIndex:03d}_left.png"), leftFrame)
            cv2.imwrite(str(self.output_folder / f"img_{self.frameIndex:03d}_right.png"), rightFrame)
        disparityPixels = getFrame(self.disparityQueue)
        leftFrame = cv2.cvtColor (leftFrame, cv2.COLOR_GRAY2RGB)
        rightFrame = cv2.cvtColor (rightFrame, cv2.COLOR_GRAY2RGB)
        self.frameIndex += 1
        return visualizer.InputPair(leftFrame, rightFrame, self.calibration, "OAK-D Camera", disparityPixels)

class StereoFromOakInputSource(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("Input Source", "Stereo computed by the input source", {}, config)

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        # The disparity is aligned to the right image with OAK-D
        return StereoOutput(input.input_disparity, input.right_image, 0.0)
        


import time

import numpy as np
import cv2

from .methods import Config, StereoMethod, IntParameter, EnumParameter, InputPair, StereoOutput

def odd_only(x):
    return x if x % 2 == 1 else x+1

def multiple_of_16(x):
    return max(16, x//16 * 16)

# Default parameters taken from 
# https://github.com/opencv/opencv/blob/4.x/samples/cpp/stereo_match.cpp
class StereoBM(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("OpenCV BM", "OpenCV Simple Block Matching", {}, config)
        self.reset_defaults()

    def reset_defaults(self):
        # For more details:
        # https://learnopencv.com/depth-perception-using-stereo-camera-python-c/
        self.parameters.update ({
            "Num Disparities": IntParameter("Number of disparities (pixels)", 128, 16, 640, to_valid=multiple_of_16),
            "Block Size": IntParameter("Kernel size for block matching (odd)", 9, 3, 63, to_valid=odd_only),
            "TextureThreshold": IntParameter("Minimum SAD to consider the texture sufficient", 10, 0, 100),
            "Uniqueness Ratio": IntParameter("How unique the match each for each pixel", 15, 0, 100),
            "SpeckleWindowSize": IntParameter("Speckle window size in pixels (filter small objects). 0 to disable.", 100, 0, 1000),
            "SpeckleRange": IntParameter("Speckle range (max diff within a window)", 32, 0, 64),
            "Disp12MaxDiff": IntParameter("Maximum allowable difference in the right-left check", 1, 0, 64),
            "PreFilterCap": IntParameter("Max pre-filter output", 31, 1, 63),
            "PreFilterSize": IntParameter("Pre-filter size (odd)", 9, 5, 255, to_valid=odd_only),
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        left_image, right_image = input.left_image, input.right_image
        block_size = self.parameters['Block Size'].value
        if block_size % 2 == 0:
            block_size += 1
        stereoBM = cv2.StereoBM_create(numDisparities=self.parameters['Num Disparities'].value, 
                                       blockSize=block_size)
        stereoBM.setTextureThreshold(self.parameters['TextureThreshold'].value)
        stereoBM.setUniquenessRatio(self.parameters['Uniqueness Ratio'].value)
        stereoBM.setSpeckleWindowSize(self.parameters['SpeckleWindowSize'].value)
        stereoBM.setSpeckleRange(self.parameters['SpeckleRange'].value)
        stereoBM.setDisp12MaxDiff(self.parameters['Disp12MaxDiff'].value)
        stereoBM.setPreFilterCap(self.parameters['PreFilterCap'].value)
        stereoBM.setPreFilterSize(self.parameters['PreFilterSize'].value)

        gray_left = cv2.cvtColor(left_image, cv2.COLOR_BGR2GRAY, left_image)
        gray_right = cv2.cvtColor(right_image, cv2.COLOR_BGR2GRAY, right_image)
        # OpenCV returns 16x the disparity in pixels
        start = time.time()
        disparity = stereoBM.compute(gray_left, gray_right) / np.float32(16.0)
        return StereoOutput(disparity, input.left_image, time.time()-start)

class StereoSGBM(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("OpenCV SGBM", "OpenCV Semi-Global Block Matching", {}, config)
        self.reset_defaults ()

    def reset_defaults(self):
        nchannels = 1
        # For more details:
        # https://learnopencv.com/depth-perception-using-stereo-camera-python-c/
        self.parameters.update ({
            "Num Disparities": IntParameter("Number of disparities (pixels)", 128, 2, 640),
            "Block Size": IntParameter("Kernel size for block matching (odd)", 3, 3, 63, to_valid=odd_only),
            
            "Mode": EnumParameter("Set it to StereoSGBM::MODE_HH to run the full-scale two-pass dynamic programming algorithm. It will consume O(W*H*numDisparities) bytes, which is large for 640x480 stereo and huge for HD-size pictures. By default, it is set to false .",
                                  0, ["MODE_SGBM", "MODE_HH", "MODE_SGBM_3WAY", "MODE_HH4"]),
            
            
            "P1": IntParameter("Penalty Cost (default=8*NChannels*BlockSize)", 8*nchannels*3*3, 0, 2000),
            "P2": IntParameter("Penalty Cost. Must be > P1 (default=32*NChannels*BlockSize).", 32*nchannels*3*3, 0, 2000),
            
            "Uniqueness Ratio": IntParameter("How unique the match each for each pixel", 10, 0, 100),
            "SpeckleWindowSize": IntParameter("Speckle window size in pixels (filter small objects). 0 to disable.", 100, 0, 1000),
            "SpeckleRange": IntParameter("Speckle range (max diff within a window)", 32, 0, 64),
            
            "Disp12MaxDiff": IntParameter("Maximum allowable difference in the right-left check", 1, 0, 64),
            "PreFilterCap": IntParameter("Max pre-filter output", 63, 1, 128),
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        left_image, right_image = input.left_image, input.right_image
        stereoSGBM = cv2.StereoSGBM_create(numDisparities=self.parameters['Num Disparities'].value, 
                                         blockSize=self.parameters['Block Size'].value)

        stereoSGBM.setMode(self.parameters['Mode'].index)
        stereoSGBM.setP1(self.parameters['P1'].value)
        stereoSGBM.setP2(self.parameters['P2'].value)
        stereoSGBM.setPreFilterCap(self.parameters['PreFilterCap'].value)
        stereoSGBM.setUniquenessRatio(self.parameters['Uniqueness Ratio'].value)
        stereoSGBM.setSpeckleWindowSize(self.parameters['SpeckleWindowSize'].value)
        stereoSGBM.setSpeckleRange(self.parameters['SpeckleRange'].value)
        stereoSGBM.setDisp12MaxDiff(self.parameters['Disp12MaxDiff'].value)

        gray_left = cv2.cvtColor(left_image, cv2.COLOR_BGR2GRAY, left_image)
        gray_right = cv2.cvtColor(right_image, cv2.COLOR_BGR2GRAY, right_image)
        # OpenCV returns 16x the disparity in pixels
        start = time.time()
        disparity = stereoSGBM.compute(gray_left, gray_right) / np.float32(16.0)
        return StereoOutput(disparity, input.left_image, time.time()-start)



from pathlib import Path
import shutil
import time
from dataclasses import dataclass
import urllib.request
import tempfile
import sys

import onnxruntime

import cv2
import numpy as np

from .methods import EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "chang-realtime-stereo-cpu-1280x720.onnx": "",
    "chang-realtime-stereo-cpu-160x128.onnx": "",
    "chang-realtime-stereo-cpu-320x240.onnx": "",
    "chang-realtime-stereo-cpu-640x480.onnx": "",
}

# Adapted from https://github.com/ibaiGorordo/ONNX-CREStereo-Depth-Estimation
# https://github.com/PINTO0309/PINTO_model_zoo/tree/main/284_CREStereo
# IMPORTANT: these ONNX are not working, keeping in case things improve later on.
class ChangRealtimeStereoOnnx(StereoMethod):
    def __init__(self):
        super().__init__("Chang Real-time Onnx", "Attention-Aware Feature Aggregation for Real-time Stereo Matching on Edge Devices (ACCV 2020).", {})
        self.reset_defaults()

        self._loaded_session = None
        self._loaded_model_path = None

    def reset_defaults(self):
        self.parameters.update ({
            "Shape": EnumParameter("Processed image size", 1, ["160x128", "320x240", "640x480", "1280x720"])
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        if not models_path.exists():
            models_path.mkdir(parents=True, exist_ok=True)
        
        cols, rows = self.parameters["Shape"].value.split('x')
        cols, rows = int(cols), int(rows)
        self.target_size = (cols, rows)

        model_path = models_path / f'chang-realtime-stereo-cpu-{cols}x{rows}.onnx'
        self._load_model (model_path)

        left_tensor = self._preprocess_input(input.left_image)
        right_tensor = self._preprocess_input(input.right_image)

        start = time.time()
        model_inputs = self._loaded_session.get_inputs()
        model_outputs = self._loaded_session.get_outputs()
        input_names = [model_inputs[i].name for i in range(len(model_inputs))]
        output_names = [model_outputs[i].name for i in range(len(model_outputs))]
        outputs = self._loaded_session.run(['disparity'], {'left': left_tensor,
                                                           'right': right_tensor})
        elapsed_time = time.time() - start

        disparity_map = self._process_output(outputs)
        if disparity_map.shape[:2] != input.left_image.shape[:2]:
            disparity_map = cv2.resize (disparity_map, (input.left_image.shape[1], input.left_image.shape[0]), cv2.INTER_NEAREST)
            x_scale = input.left_image.shape[1] / float(cols)
            disparity_map *= np.float32(x_scale)

        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _preprocess_input (self, img: np.ndarray):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, self.target_size, cv2.INTER_AREA)
        imagenet_stats = {'mean': np.array([0.485, 0.456, 0.406]), 'std': np.array([0.229, 0.224, 0.225])}
        img = (img.astype(np.float32) / 255.0) - imagenet_stats['mean'] / imagenet_stats['std']
        img = img.transpose(2, 0, 1) # C,H,W instead of H,W,C
        img = img[np.newaxis, :, :, :] # add batch dimension
        return img

    def _process_output(self, outputs):
        disparity_map = outputs[0].permute(1,2,0)
        return disparity_map

    def _load_model(self, model_path: Path):
        if (self._loaded_model_path == model_path):
            return
        
        if not model_path.exists():
            utils.download_model (urls[model_path.name], model_path)

        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self._loaded_session = onnxruntime.InferenceSession(str(model_path), providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])



from pathlib import Path
import shutil
import time
from dataclasses import dataclass
import urllib.request
import tempfile
import sys

import torch
from torchvision import transforms

import cv2
import numpy as np

from .methods import Calibration, Config, EnumParameter, StereoMethod, InputPair, StereoOutput
from . import utils

urls = {
    "dist-depth-256x256.scripted.pt": "https://github.com/nburrus/stereodemo/releases/download/v0.1-distdepth/dist-depth-256x256.scripted.pt",
}

# https://github.com/facebookresearch/DistDepth
# Exported via torch tracing by tweaking the original demo.py.
# Changes here: https://github.com/nburrus/DistDepth/commit/fde3b427ef2ff31c34f08e99c51c8e6a2427b720
class DistDepth(StereoMethod):
    def __init__(self, config: Config):
        super().__init__("[Monocular] DistDepth (CVPR 2022)",
                         "Toward Practical Monocular Indoor Depth Estimation.",
                         {},
                         config)
        self.reset_defaults()

        self.net = None
        self._loaded_model_path = None

    def reset_defaults(self):
        self.parameters.update ({
            # "Device": EnumParameter("Device", 0, ["CPU", "CUDA"]),
            # For some reason it crashes with CUDA on my machine, disabling for now.
            "Device": EnumParameter("Device", 0, ["CPU"]),
        })

    def compute_disparity(self, input: InputPair) -> StereoOutput:
        # The pre-trained model is for 256x256. Their demo script resizes
        # all input images to that.
        self.target_size = (256, 256)
        device = torch.device('cuda') if self.parameters["Device"].value == 'CUDA' else 'cpu'
        
        model_path = self.config.models_path / f'dist-depth-256x256.scripted.pt'
        self._load_model (model_path)

        # raw_img can stay in BGR
        raw_img = np.transpose(input.left_image, (2, 0, 1))
        input_image = torch.from_numpy(raw_img).float().to(device)
        input_image = (input_image / 255.0).unsqueeze(0)
        input_image = torch.nn.functional.interpolate(
            input_image, (256, 256), mode="bilinear", align_corners=False
        )

        net = self.net.to(device)

        start = time.time()
        with torch.no_grad():
            outputs = net(input_image.to(device))
        elapsed_time = time.time() - start

        disparity_map = self._process_output(outputs, input.calibration)
        if disparity_map.shape[:2] != input.left_image.shape[:2]:
            disparity_map = cv2.resize (disparity_map, (input.left_image.shape[1], input.left_image.shape[0]), cv2.INTER_NEAREST)
            # not need to scale, the disparity values were already for the input full resolution calibration.

        return StereoOutput(disparity_map, input.left_image, elapsed_time)

    def _process_output(self, outputs, calib: Calibration):
        depth_meters = outputs[0].detach().squeeze(0).cpu().numpy()
        # The model directly gives a depth map in meters. Let's convert it
        # to disparity to fit in the stereo display.
        disparity_map = StereoMethod.disparity_from_depth_meters(depth_meters, calib)
        return disparity_map

    def _load_model(self, model_path: Path):
        if (self._loaded_model_path == model_path):
            return
        
        if not model_path.exists():
            utils.download_model (urls[model_path.name], model_path)

        assert Path(model_path).exists()
        self._loaded_model_path = model_path
        self.net = torch.jit.load(model_path)
        self.net.cpu ()
        self.net.eval ()
